@@ -141,9 +141,6 @@ class Hdf5Iterator:
         else:
             arrays = dataset[self._current_start_entry : self._current_stop_entry]
 
-        if self.shuffle:
-            np.random.shuffle(arrays)
-
         return arrays
 
     def __next__(self) -> tuple[Any, dict[str, Any]]:
