@@ -28,6 +28,10 @@ WeightedBatchType = tuple[
     torch.Tensor,
     dict[str, Any],
 ]
+FullWeightedBatchType = tuple[
+    dict[str, tuple[torch.Tensor, torch.Tensor | None, torch.Tensor | None, torch.Tensor | None]],
+    dict[str, Any],
+]
 
 
 @dataclass(frozen=True, slots=True, repr=False, eq=False)
@@ -165,6 +169,16 @@ class WeightedBatch:
         if isinstance(self._y_aux, np.ndarray):
             self._y_aux = torch.from_numpy(self._y_aux)
 
+        n_samples = self._X.shape[0]
+        if self._y is not None and self._y.shape[0] != n_samples:
+            raise RuntimeError("X and y must have the same first dimension!")
+
+        if self._w is not None and self._w.shape[0] != n_samples:
+            raise RuntimeError("X and w must have the same first dimension!")
+
+        if self._y_aux is not None and self._y_aux.shape[0] != n_samples:
+            raise RuntimeError("X and y_aux must have the same first dimension!")
+
     @property
     def X(self) -> torch.Tensor:
         return self._X  # type: ignore
@@ -185,8 +199,7 @@ class WeightedBatch:
     def feature_shape(self) -> tuple[int, ...]:
         return tuple(self._X.shape)
 
-    @property
-    def batch(self) -> tuple[torch.Tensor, torch.Tensor | None, torch.Tensor | None, torch.Tensor | None]:
+    def make_batch(self) -> tuple[torch.Tensor, torch.Tensor | None, torch.Tensor | None, torch.Tensor | None]:
         return self._X, self._y, self._w, self._y_aux  # type: ignore
 
     def mask(self, mask: np.ndarray | torch.Tensor) -> WeightedBatch:
@@ -283,6 +296,14 @@ class WeightedDatasetBatch:
 
         return self.__class__(_masked_datasets)
 
+    def make_batch(
+        self,
+    ) -> dict[str, tuple[torch.Tensor, torch.Tensor | None, torch.Tensor | None, torch.Tensor | None]]:
+        batch = {}
+        for key, ds_batch in self._datasets.items():
+            batch[key] = ds_batch.make_batch()
+        return batch
+
     def __repr__(self) -> str:
         return f"WeightedDatasetBatch({self.keys()})"
 
@@ -393,7 +414,7 @@ class MLHdf5Iterator(Hdf5Iterator):
         postfix: str | None = None,
         extra_hash: str = "",
     ) -> tuple[NumericalFeatureScaler | None, CategoricalFeatureScaler | None]:
-        if scaler_path is None and scaler_type is None:
+        if scaler_path is None or scaler_type is None:
             return None, None
 
         numer_feature_scaler: NumericalFeatureScaler | None
@@ -422,7 +443,7 @@ class MLHdf5Iterator(Hdf5Iterator):
 
         return numer_feature_scaler, categ_feature_scaler
 
-    def _scale(self, X: np.ndarray, dataset_name: str, is_numer: bool) -> np.ndarray:
+    def _scale(self, X: np.ndarray, dataset_name: str, is_numer: bool, scale_categ_padding: bool = True) -> np.ndarray:
         scalers: list[NumericalFeatureScaler | None] | list[CategoricalFeatureScaler | None]
 
         if is_numer:
@@ -432,6 +453,9 @@ class MLHdf5Iterator(Hdf5Iterator):
 
         n_objects = self.selection[dataset_name].n_objects
 
+        if len(scalers) != n_objects:
+            raise RuntimeError(f"Scaler list length mismatch for dataset {dataset_name}.")
+
         if n_objects == 1:
             scaler = scalers[0]
             if scaler is not None:
@@ -443,7 +467,9 @@ class MLHdf5Iterator(Hdf5Iterator):
                     X_i = X[:, i, :]
                     pad_value = self.selection[dataset_name].pad_value
 
-                    if pad_value is not None:
+                    if pad_value is not None and scale_categ_padding and not is_numer:
+                        X_objects.append(scaler.transform(X_i))
+                    elif pad_value is not None:
                         mask = padding_mask(X_i, pad_value)
 
                         X_full = np.full_like(X_i, pad_value)
@@ -504,16 +530,16 @@ class MLHdf5Iterator(Hdf5Iterator):
 
             if X_numer_dataset_stack is not None and X_categ_dataset_stack is not None:
                 X_dataset_stack = np.concatenate([X_numer_dataset_stack, X_categ_dataset_stack], axis=-1)
-                numer_idx = np.arange(numer_shape[1])
-                categ_idx = np.arange(numer_shape[1], numer_shape[1] + categ_shape[1])
+                numer_idx = np.arange(numer_shape[-1])
+                categ_idx = np.arange(numer_shape[-1], numer_shape[-1] + categ_shape[-1])
 
             elif X_numer_dataset_stack is not None:
                 X_dataset_stack = X_numer_dataset_stack
-                categ_idx, numer_idx = None, np.arange(X_numer_dataset_stack.shape[1])
+                categ_idx, numer_idx = None, np.arange(numer_shape[-1])
 
             elif X_categ_dataset_stack is not None:
                 X_dataset_stack = X_categ_dataset_stack
-                categ_idx, numer_idx = np.arange(X_categ_dataset_stack.shape[1]), None
+                categ_idx, numer_idx = np.arange(categ_shape[-1]), None
 
             else:
                 raise RuntimeError("Both X_numer and X_categ are None!")
@@ -629,10 +655,24 @@ def events_collate_fn(batch: Any) -> WeightedBatchType:
     return ds.X, ds.y, ds.w, ds.y_aux, reports
 
 
-def default_setup_func(stacked_datasets: StackedDatasets, ml_iterator: MLHdf5Iterator) -> WeightedDatasetBatch:
+def full_collate_fn(batch: Any) -> Any:
+    datasets_batch, reports = batch[0].make_batch(), batch[1]
+    return datasets_batch, reports
+
+
+def default_setup_func(stacked_datasets: StackedDatasets, ml_iterator: MLHdf5Iterator) -> WeightedDatasetBatch | None:
     weighted_datase_batch = WeightedDatasetBatch()
 
-    for ds_name, ds in stacked_datasets.items():
+    dataset_keys = stacked_datasets.keys()
+    if "events" in dataset_keys:
+        dataset_keys.remove("events")
+        dataset_keys.insert(0, "events")
+
+    mask_unmapped: np.ndarray | None = None
+
+    for ds_name in dataset_keys:
+        ds = stacked_datasets[ds_name]
+
         X = ds.X.astype(np.float32)
         y = ds.get_extra("label_type", None)
         w = ds.get_extra("weights", None)
@@ -643,18 +683,14 @@ def default_setup_func(stacked_datasets: StackedDatasets, ml_iterator: MLHdf5Ite
             if remap_labels is not None:
                 max_label = ml_iterator.dataset_kwargs["max_label"]
                 y, mask_unmapped = remap_labels_lookup(y, max_label, remap_labels)
-                X = X[mask_unmapped]
-                if w is not None:
-                    w = w[mask_unmapped]
 
-        if ml_iterator.shuffle:
-            shuffle_idx = np.random.permutation(len(X))
-            X[:] = X[shuffle_idx]
+            if y.shape[0] == 0:
+                return None
 
-            if y is not None:
-                y[:] = y[shuffle_idx]
+        if mask_unmapped is not None:
+            X = X[mask_unmapped]
             if w is not None:
-                w[:] = w[shuffle_idx]
+                w = w[mask_unmapped]
 
         weighted_datase_batch[ds_name] = WeightedBatch(X, y, w, y_aux)
 
